# Copyright (C) 2025 Maxwell J. Campbell
"""MDSASA-Bolt: Solvent Accessible Surface Area Analysis for MD Trajectories."""

from .analysis import SASAAnalysis

__version__ = "0.1.0"
__all__ = ["SASAAnalysis"]
